# Copyright 2023-2024 GÉANT Vereniging.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import re
import time
from collections.abc import Callable
from unittest.mock import patch

import responses
from fastapi import status
from fastapi.testclient import TestClient

TEST_CALLBACK_URL = "https://fqdn.abc.xyz/api/resume"


@responses.activate
def test_playbook_endpoint_dict_inventory_success(client: TestClient, mocked_ansible_runner_run: Callable) -> None:
    responses.post(url=TEST_CALLBACK_URL, status=status.HTTP_200_OK)

    params = {
        "playbook_name": "placeholder.yaml",
        "callback": TEST_CALLBACK_URL,
        "inventory": {
            "_meta": {"vars": {"host1.local": {"foo": "bar"}, "host2.local": {"hello": "world"}}},
            "all": {"hosts": {"host1.local": None, "host2.local": None}},
        },
        "extra_vars": {"dry_run": True, "commit_comment": "I am a robot!"},
    }

    with patch("lso.playbook.ansible_runner.run", new=mocked_ansible_runner_run) as _:
        rv = client.post("/api/playbook/", json=params)
        assert rv.status_code == status.HTTP_201_CREATED
        response = rv.json()
        # wait one second for the run thread to finish
        time.sleep(1)

    assert isinstance(response, dict)
    assert isinstance(response["job_id"], str)
    responses.assert_call_count(TEST_CALLBACK_URL, 1)


@responses.activate
def test_playbook_endpoint_str_inventory_success(client: TestClient, mocked_ansible_runner_run: Callable) -> None:
    responses.post(url=TEST_CALLBACK_URL, status=status.HTTP_200_OK)

    params = {
        "playbook_name": "placeholder.yaml",
        "callback": TEST_CALLBACK_URL,
        "inventory": {"all": {"hosts": "host1.local\nhost2.local\nhost3.local"}},
    }

    with patch("lso.playbook.ansible_runner.run", new=mocked_ansible_runner_run) as _:
        rv = client.post("/api/playbook/", json=params)
        assert rv.status_code == status.HTTP_201_CREATED
        response = rv.json()
        # wait one second for the run thread to finish
        time.sleep(1)

    assert isinstance(response, dict)
    assert isinstance(response["job_id"], str)
    responses.assert_call_count(TEST_CALLBACK_URL, 1)


@responses.activate
def test_playbook_endpoint_invalid_host_vars(client: TestClient, mocked_ansible_runner_run: Callable) -> None:
    params = {
        "playbook_name": "placeholder.yaml",
        "callback": TEST_CALLBACK_URL,
        "inventory": {
            "_meta": {"host_vars": {"host1.local": {"foo": "bar"}, "host2.local": {"hello": "world"}}},
            "all": {"hosts": "host1.local\nhost2.local\nhost3.local"},
        },
    }

    with patch("lso.playbook.ansible_runner.run", new=mocked_ansible_runner_run) as _:
        rv = client.post("/api/playbook/", json=params)
        assert rv.status_code == status.HTTP_422_UNPROCESSABLE_ENTITY
        response = rv.json()
        # wait one second for the run thread to finish
        time.sleep(1)

    assert isinstance(response, dict)
    assert response["detail"] == [
        '[WARNING]: Skipping unexpected key (host_vars) in group (_meta), only "vars",\n',
        '"children" and "hosts" are valid\n',
    ]
    responses.assert_call_count(TEST_CALLBACK_URL, 0)


@responses.activate
def test_playbook_endpoint_invalid_hosts(client: TestClient, mocked_ansible_runner_run: Callable) -> None:
    params = {
        "playbook_name": "placeholder.yaml",
        "callback": TEST_CALLBACK_URL,
        "inventory": {
            "_meta": {"vars": {"host1.local": {"foo": "bar"}}},
            "all": {"hosts": ["host1.local", "host2.local", "host3.local"]},
        },
    }

    with patch("lso.playbook.ansible_runner.run", new=mocked_ansible_runner_run) as _:
        rv = client.post("/api/playbook/", json=params)
        assert rv.status_code == status.HTTP_422_UNPROCESSABLE_ENTITY
        response = rv.json()
        # wait one second for the run thread to finish
        time.sleep(1)

    assert isinstance(response, dict)
    assert 'Invalid "hosts" entry for "all" group' in re.sub("\n", " ", "".join(response["detail"]))
    responses.assert_call_count(TEST_CALLBACK_URL, 0)



# Copyright 2023-2024 GÉANT Vereniging.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from importlib import metadata

import jsonschema
import responses
from fastapi import status
from starlette.testclient import TestClient

from lso.routes.default import API_VERSION, Version


@responses.activate
def test_ip_trunk_modification(client: TestClient) -> None:
    rv = client.get("/api/version/")
    assert rv.status_code == status.HTTP_200_OK, rv.text
    response = rv.json()

    jsonschema.validate(response, Version.model_json_schema())

    assert response["api"] == API_VERSION, response["api"]
    assert response["module"] == metadata.version("orchestrator-lso"), response["module"]



# Copyright 2023-2024 GÉANT Vereniging.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""A module for loading configuration data, including a configuration schema that data is validated against.

Data is loaded from a file, the location of which may be specified when using :func:`load_from_file`.
Configuration file location can also be loaded from environment variable ``$SETTINGS_FILENAME``, which is default
 behaviour in :func:`load`.
"""

import json
import os
from pathlib import Path

import jsonschema
from pydantic import BaseModel

CONFIG_SCHEMA = {
    "$schema": "http://json-schema.org/draft-07/schema#",
    "type": "object",
    "properties": {"ansible_playbooks_root_dir": {"type": "string"}},
    "required": ["ansible_playbooks_root_dir"],
    "additionalProperties": False,
}
DEFAULT_REQUEST_TIMEOUT = 10


class Config(BaseModel):
    """Simple Configuration class.

    Contains the root directory at which Ansible playbooks are present.
    """

    ansible_playbooks_root_dir: str


def load_from_file(file: Path) -> Config:
    """Load, validate and return configuration parameters.

    Input is validated against this JSON schema:

    .. asjson:: lso.config.CONFIG_SCHEMA

    :param file: :class:`Path` object that produces the configuration file.
    :return: a dict containing the parsed configuration parameters.
    """
    config = json.loads(file.read_text())
    jsonschema.validate(config, CONFIG_SCHEMA)
    return Config(**config)


def load() -> Config:
    """Load a configuration file, located at the path specified in the environment variable ``$SETTINGS_FILENAME``.

    Loading and validating the file is performed by :func:`load_from_file`.

    :return: a dict containing the parsed configuration parameters
    """
    assert "SETTINGS_FILENAME" in os.environ, "Environment variable SETTINGS_FILENAME not set"  # noqa: S101
    return load_from_file(Path(os.environ["SETTINGS_FILENAME"]))



# Copyright 2023-2024 GÉANT Vereniging.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Set of tests that verify correct config is accepted and incorrect config is not."""

import json
import os
import tempfile
from pathlib import Path

import jsonschema
import pytest

from lso import config


def test_validate_testenv_config(data_config_filename: str) -> None:
    """Load a configuration from a file.

    :param data_config_filename: Configuration file pytest fixture
    """
    os.environ["SETTINGS_FILENAME"] = data_config_filename
    params = config.load()
    assert params


@pytest.mark.parametrize(
    "bad_config", [{"name": "bad version", "version": 123}, {"name": "missing version"}, {"version": "missing name"}]
)
def test_bad_config(bad_config: dict) -> None:
    with tempfile.NamedTemporaryFile(mode="w") as file:
        Path(file.name).write_text(json.dumps(bad_config))
        with pytest.raises(jsonschema.ValidationError):
            config.load_from_file(Path(file.name))



# Copyright 2023-2024 GÉANT Vereniging.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""The API endpoint from which Ansible playbooks can be executed."""

import json
import tempfile
from contextlib import redirect_stderr
from io import StringIO
from typing import Annotated, Any

from ansible.inventory.manager import InventoryManager
from ansible.parsing.dataloader import DataLoader
from fastapi import APIRouter, HTTPException, status
from fastapi.responses import JSONResponse
from pydantic import AfterValidator, BaseModel, HttpUrl

from lso.playbook import get_playbook_path, run_playbook

router = APIRouter()


def _inventory_validator(inventory: dict[str, Any] | str) -> dict[str, Any] | str:
    """Validate the format of the provided inventory by trying to parse it.

    If an inventory can't be parsed without warnings or errors, these are returned to the user by means of an HTTP
    status 422 for 'unprocessable entity'.
    """
    loader = DataLoader()
    output = StringIO()
    with tempfile.NamedTemporaryFile(mode="w+") as temp_inv, redirect_stderr(output):
        json.dump(inventory, temp_inv, ensure_ascii=False)
        temp_inv.flush()

        inventory_manager = InventoryManager(loader=loader, sources=[temp_inv.name], parse=True)
        inventory_manager.parse_source(temp_inv.name)

    output.seek(0)
    error_messages = output.readlines()
    if error_messages:
        raise HTTPException(status_code=status.HTTP_422_UNPROCESSABLE_ENTITY, detail=error_messages)

    return inventory


PlaybookInventory = Annotated[dict[str, Any] | str, AfterValidator(_inventory_validator)]


class PlaybookRunParams(BaseModel):
    """Parameters for executing an Ansible playbook."""

    #: The filename of a playbook that's executed. It should be present inside the directory defined in the
    #: configuration option ``ansible_playbooks_root_dir``.
    playbook_name: str
    #: The address where LSO should call back to upon completion.
    callback: HttpUrl
    #: The inventory to run the playbook against. This inventory can also include any host vars, if needed. When
    #: including host vars, it should be a dictionary. Can be a simple string containing hostnames when no host vars are
    #: needed. In the latter case, multiple hosts should be separated with a ``\n`` newline character only.
    inventory: PlaybookInventory
    #: Extra variables that should get passed to the playbook. This includes any required configuration objects
    #: from the workflow orchestrator, commit comments, whether this execution should be a dry run, a trouble ticket
    #: number, etc. Which extra vars are required solely depends on what inputs the playbook requires.
    extra_vars: dict[str, Any] = {}


@router.post("/")
def run_playbook_endpoint(params: PlaybookRunParams) -> JSONResponse:
    """Launch an Ansible playbook to modify or deploy a subscription instance.

    The response will contain either a job ID, or error information.

    :param PlaybookRunParams params: Parameters for executing a playbook.
    :return JSONResponse: Response from the Ansible runner, including a run ID.
    """
    return run_playbook(
        playbook_path=get_playbook_path(params.playbook_name),
        extra_vars=params.extra_vars,
        inventory=params.inventory,
        callback=params.callback,
    )



# Copyright 2023-2024 GÉANT Vereniging.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Module that gathers common API responses and data models."""

import logging
import threading
import uuid
from pathlib import Path
from typing import Any

import ansible_runner
import requests
from fastapi import status
from fastapi.responses import JSONResponse
from pydantic import HttpUrl

from lso import config
from lso.config import DEFAULT_REQUEST_TIMEOUT

logger = logging.getLogger(__name__)


def get_playbook_path(playbook_name: str) -> Path:
    """Get the path of a playbook on the local filesystem."""
    config_params = config.load()
    return Path(config_params.ansible_playbooks_root_dir) / playbook_name


def playbook_launch_success(job_id: str) -> JSONResponse:
    """Return a :class:`PlaybookLaunchResponse` for the successful start of a playbook execution.

    :return JSONResponse: A playbook launch response that's successful.
    """
    return JSONResponse(content={"job_id": job_id}, status_code=status.HTTP_201_CREATED)


def playbook_launch_error(reason: str, status_code: int = status.HTTP_400_BAD_REQUEST) -> JSONResponse:
    """Return a :class:`PlaybookLaunchResponse` for the erroneous start of a playbook execution.

    :param str reason: The reason why a request has failed.
    :param status status_code: The HTTP status code that should be associated with this request. Defaults to HTTP 400:
                               Bad request.
    :return JSONResponse: A playbook launch response that's unsuccessful.
    """
    return JSONResponse(content={"error": reason}, status_code=status_code)


def _run_playbook_proc(
    job_id: str, playbook_path: str, extra_vars: dict, inventory: dict[str, Any] | str, callback: str
) -> None:
    """Run a playbook, internal function.

    :param str job_id: Identifier of the job that's executed.
    :param str playbook_path: Ansible playbook to be executed.
    :param dict extra_vars: Extra variables passed to the Ansible playbook.
    :param str callback: Callback URL to return output to when execution is completed.
    :param dict[str, Any] | str inventory: Ansible inventory to run the playbook against.
    """
    ansible_playbook_run = ansible_runner.run(playbook=playbook_path, inventory=inventory, extravars=extra_vars)

    payload = {
        "status": ansible_playbook_run.status,
        "job_id": job_id,
        "output": ansible_playbook_run.stdout.readlines(),
        "return_code": int(ansible_playbook_run.rc),
    }

    request_result = requests.post(callback, json=payload, timeout=DEFAULT_REQUEST_TIMEOUT)
    if not status.HTTP_200_OK <= request_result.status_code < status.HTTP_300_MULTIPLE_CHOICES:
        msg = f"Callback failed: {request_result.text}"
        logger.error(msg)


def run_playbook(
    playbook_path: Path,
    extra_vars: dict[str, Any],
    inventory: dict[str, Any] | str,
    callback: HttpUrl,
) -> JSONResponse:
    """Run an Ansible playbook against a specified inventory.

    :param Path playbook_path: playbook to be executed.
    :param dict[str, Any] extra_vars: Any extra vars needed for the playbook to run.
    :param dict[str, Any] | str inventory: The inventory that the playbook is executed against.
    :param HttpUrl callback: Callback URL where the playbook should send a status update when execution is completed.
                             This is used for workflow-orchestrator to continue with the next step in a workflow.
    :return: Result of playbook launch, this could either be successful or unsuccessful.
    :rtype: :class:`fastapi.responses.JSONResponse`
    """
    if not Path.exists(playbook_path):
        msg = f"Filename '{playbook_path}' does not exist."
        return playbook_launch_error(reason=msg, status_code=status.HTTP_404_NOT_FOUND)

    if not ansible_runner.utils.isinventory(inventory):
        msg = "Invalid inventory provided. Should be a string, or JSON object."
        return playbook_launch_error(reason=msg, status_code=status.HTTP_400_BAD_REQUEST)

    job_id = str(uuid.uuid4())
    thread = threading.Thread(
        target=_run_playbook_proc,
        kwargs={
            "job_id": job_id,
            "playbook_path": str(playbook_path),
            "inventory": inventory,
            "extra_vars": extra_vars,
            "callback": callback,
        },
    )
    thread.start()

    return playbook_launch_success(job_id=job_id)



# Copyright 2023-2024 GÉANT Vereniging.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import os
import tempfile
from collections.abc import Callable, Generator
from io import StringIO
from pathlib import Path
from typing import Any

import pytest
from faker import Faker
from fastapi.testclient import TestClient

import lso


@pytest.fixture
def mocked_ansible_runner_run() -> Callable:
    class Runner:
        def __init__(self) -> None:
            self.status = "success"
            self.rc = 0
            self.stdout = StringIO("[{'step one': 'results'}, {'step two': 2}]")

    def run(*args: Any, **kwargs: Any) -> Runner:  # noqa: ARG001
        return Runner()

    return run


@pytest.fixture(scope="session")
def configuration_data() -> dict[str, str]:
    """Start the server with valid configuration data."""
    with tempfile.TemporaryDirectory() as tempdir:
        # Create required YAML files for the unit tests
        (Path(tempdir) / "placeholder.yaml").touch()

        yield {"ansible_playbooks_root_dir": tempdir}


@pytest.fixture(scope="session")
def data_config_filename(configuration_data: dict[str, str]) -> Generator[str, Any, None]:
    """Fixture that will yield a filename that contains a valid configuration.

    :return: Path to valid configuration file
    """
    with tempfile.NamedTemporaryFile(mode="w") as file:
        file.write(json.dumps(configuration_data))
        file.flush()
        yield file.name


@pytest.fixture(scope="session")
def client(data_config_filename: str) -> TestClient:
    """Return a client that can be used to test the server."""
    os.environ["SETTINGS_FILENAME"] = data_config_filename
    app = lso.create_app()
    return TestClient(app)  # wait here until calling context ends


@pytest.fixture(scope="session")
def faker() -> Faker:
    return Faker(locale="en_GB")
